
#include <linux/kernel.h>
#include <linux/module.h>

#include <linux/fs.h>
#include <linux/delay.h>
#include <asm/io.h>

#define DEVICE_NAME   "analog_out"
#define DAC_MAJOR     177

static int base = 0x240;

static int Device_Open = 0;

static int device_open(struct inode *inode, struct file *file)
{
  if (Device_Open)
    return -EBUSY;

  Device_Open++;
  MOD_INC_USE_COUNT;

  return 0;
}

static int device_release(struct inode *inode, struct file *file)
{
  Device_Open--;
  MOD_DEC_USE_COUNT;
  return 0;
}

#define set_mode(ctr, mode) ({ b = ((mode << 1) & 0x0E) | 0x030| ((ctr << 6) & 0x0C0); outb(b, base+11); })
#define load(ctr, amt) ({ outb(amt & 0xff, base+8+ctr); outb((amt >> 8) & 0xff, base+8+ctr); })

static ssize_t device_write(struct file *file, const char *buffer, size_t length, loff_t *offset)
{
	int chan, flags;
	int b, i;
	
//	if (length != 6 && length != 1024*2*3)
//		return -EINVAL;

	if (length == 6)
	{
		/* direct output mode */
		for (chan = 1; chan <= 3; chan++) {
			flags = (chan << 4) | chan;
			outb(flags, base+2);  /* start and stop channels */
			outb(0, base+3);      /* clear start bit */
			inb(base+1);          /* reset fifo */
			b = ((unsigned short *)buffer)[chan-1] << 4;
			outb((b >> 8) & 0xff, base+1);  /* MSB */
			outb(b & 0xff, base+0);         /* LSB */
		}
	}
	else
	{
			/* timed FIFO output mode */

			set_mode(2, 2);  /* stop pacer clock on all channels */
			set_mode(1, 2);
			set_mode(0, 2);
			inb(base+1);          /* clear FIFO */
			outb(0, base+3);      /* disable interrupts */
			inb(base+0);          /* reset interrupt */
			outb((3 << 4) | 1, base+2);    /* set channel limits (1 through 3) */
			/* DAC pacer start */
			set_mode(2, 2);
			set_mode(1, 2);
			set_mode(0, 2);
			load(2, 17);
			load(1, 2);
			load(0, 2);
			/* end DAC pacer start */
			outb(3, base+3);      /* set start and simutaneous bits */
			/* stuff the FIFO */
			for (i = 0; i < length/2; i++) {
				b = ((unsigned short *)buffer)[i] << 4;
				outb((b >> 8) & 0xff, base+1);
				outb(b & 0xff, base+0);
			}
			outb_p(2, base+3);      /* clear start bit */
			udelay(100);
			outb(3, base+3);      /* set start bit */
		}

	return 0;
}

struct file_operations dac_fops = {
  NULL,   /* seek */
  NULL,   /* read */
  device_write,
  NULL,   /* readdir */
  NULL,   /* select */
  NULL,   /* ioctl */
  NULL,   /* mmap */
  device_open,
  NULL,   /* flush */
  device_release  /* close */
};

int init_module()
{
	int err;

	if ((err = register_chrdev(DAC_MAJOR, DEVICE_NAME, &dac_fops)) < 0)
	{
		printk("dac: couldn't grab major number %d", DAC_MAJOR);
		return err;
	}

	printk("DAC driver installed");

  return 0;
}


void cleanup_module()
{
	unregister_chrdev(DAC_MAJOR, DEVICE_NAME);
}

