
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <sched.h>
#include <unistd.h>
#include <math.h>

#define MAX_X            0xfff
#define MAX_Y            0xfff
#define MAX_INTENSITY    0xfff

#define SCALE_X          0.2
#define SCALE_Y          0.2
#define SCALE_INTENSITY  0.2

#define CLAMP(val, min, max) ((val) > (max) ? (max) : (val) >= (min) ? (val) : (min))

unsigned short reset_vals[3] = { 0, MAX_X/2, MAX_Y/2 };

typedef struct point_data
{
	float x, y, intensity;
} point_data;

point_data shape[] = { { 0.8, -0.5, -0.7 },
                       { 0.8,  0.5, -0.5 },
                       { 0.8,  0.5,  0.5 },
                       { 0.8, -0.5,  0.5 },
                       { 0.8, -0.5, -0.7 } };

point_data zero_point = { 0, 0, 0 };

int fd = -1;

typedef int bool;

#define true        1
#define false       0

#define sqr(x)      ((x)*(x))

float dist(point_data *start, point_data *end)
{
	return sqrt(sqr(start->x - end->x) + sqr(start->y - end->y));
}

struct timespec t = { 0, 0 };

/*
#define PAUSE_TIME  10
#define PAUSE(v) \
   { if ((v) > 0) { t.tv_nsec = (v) * PAUSE_TIME; nanosleep(&t, 0); } }
	 */
#define PAUSE(x)

#define WAIT_CONST  1
#define LINE_INCR   0.001

#define KILL_CONST  0.2

#define DAC_DELAY     5

void set_vals(point_data *s, float mod_intensity, unsigned short *vals)
{
	int v;

	/* convert the points to the monitor's coordinate system */
	v = (int)((1.0f - (s->x * SCALE_X)) * (float)(MAX_X/2.0));
	v = CLAMP(v, 0, MAX_X);
	vals[1] = v;

	v = (int)((1.0f + (s->y * SCALE_Y)) * (float)(MAX_Y/2.0));
	v = CLAMP(v, 0, MAX_Y);
	vals[2] = v;

	if (mod_intensity < 0) mod_intensity = s->intensity;
	v = (unsigned short)(mod_intensity * (float)SCALE_INTENSITY * (float)MAX_INTENSITY);
	v = CLAMP(v, 0, MAX_INTENSITY);
	vals[0] = v;
}

void draw_point(point_data *s, float mod_intensity, bool kill_intensity)
{
	static point_data current = { 0, 0, 0 };
	unsigned short vals[3];
	float travel_time, kill_time;

	set_vals(s, mod_intensity, vals);

	/* calculate the travel time */
	travel_time = dist(&current, s) * WAIT_CONST;

	/* in the case of kill_intensity, turn off the intensity part way there */
	if (kill_intensity)
	{
		kill_time = travel_time * (1 - KILL_CONST);
		travel_time -= kill_time;
	}

	/* output the data to the monitor */
	write(fd, vals, 3 * sizeof(unsigned short));

	/* pause to wait for the monitor beam to catch up */
#if 0
	PAUSE(travel_time);

	if (kill_intensity)
	{
		vals[2] = 0;
		write(fd, vals, 3 * sizeof(unsigned short));
		PAUSE(kill_time);
	}
#endif

	/* save the new position */
	current = *s;

//	PAUSE(DAC_DELAY);
}

#define TRAVEL_TIME   20
#define TRAVEL_CONST  (TRAVEL_TIME / DAC_DELAY)

void draw_line(point_data *start, point_data *end)
{
#if 1
	struct timespec req, rem;
	enum { COUNT = 1000*3 };
	unsigned short vals[COUNT];
	float i, omi, incr;
	point_data p;
	int c, b;

	c = 0;
//	for (b = 0; b < 20; b++) set_vals(start, 0, &vals[c+=3]);
//	set_vals(start, -1, &vals[c+=3]);

	incr =  1 / (dist(start, end) * TRAVEL_CONST); 
	for (i = 0; i <= 1.0f && c < COUNT-(40*3); i += incr)
	{
		omi = 1.0f - i;
		p.x = (start->x * omi) + (end->x * i);
		p.y = (start->y * omi) + (end->y * i);
		p.intensity = (start->intensity * omi) + (end->intensity * i);
		set_vals(&p, -1, &vals[c+=3]);
	}

//	set_vals(end, -1, &vals[c+=3]);
//	for (b = 0; b < 20; b++) set_vals(end, 0, &vals[c+=3]);

#if 1
	c = 500;
	for (b = 0; b < c*3; ) {
		vals[b++] = 1500;
		vals[b++] = b+1500;
		vals[b++] = b*3+500;
	}
#endif
#if 0
	for (b = 0; b < c; )
		printf("%4d %4d %4d\n", vals[b++], vals[b++], vals[b++]);

	exit(0);
#endif

	write(fd, vals, c * sizeof(unsigned short));
	req.tv_sec = 0; req.tv_nsec = 500;
	nanosleep(&req, &rem);

#else
	point_data p;
	float i, omi, incr;

	draw_point(start, 0, false);
//	PAUSE(100);
	draw_point(start, -1, false);
//	PAUSE(100);

#if 1
/*	incr = dist(start, end) * LINE_INCR;*/
	incr =  1 / (dist(start, end) * TRAVEL_CONST); 
	for (i = 0; i <= 1.0f; i += incr)
	{
		omi = 1.0f - i;
		p.x = (start->x * omi) + (end->x * i);
		p.y = (start->y * omi) + (end->y * i);
		p.intensity = (start->intensity * omi) + (end->intensity * i);
		draw_point(&p, -1, false);
	}
#else
	draw_point(end, -1, false);
#endif

	draw_point(end, 0, false);
//	PAUSE(100);
	draw_point(&zero_point, -1, false);
//	PAUSE(200);
#endif
}

void segv_handler(int sig)
{
	static int pass = 0;
	if (pass++) return;

	if (fd > 0)
	{
		write(fd, reset_vals, 3 * sizeof(unsigned short));
		close(fd);
	}
	
	exit(0);
}

int main()
{
//	struct sched_param pr;

	point_data p0, p1;
	float pos, delta;
	int point;

	/*pr.sched_priority = 1;
	sched_setscheduler(0, SCHED_RR, &pr);*/

	signal(SIGSEGV, segv_handler);
	signal(SIGTERM, segv_handler);
	signal(SIGINT, segv_handler);

//	fd = open("/dev/dac", O_WRONLY);
	fd = open("/dev/dac", O_RDWR);
	if (fd < 0) {
		printf("couldn't open /dev/dac for writing\n");
		return -1;
	}

	pos = 0;
	delta = 1;

#if 0
	{
	short vals[3];
	int which = 0;
	while (1)
	{
		vals[0] = 0xfff;
		vals[1] = 4000;
		vals[2] = 4000;
		write(fd, vals, sizeof(unsigned short));
//		which = which ? 0 : 1;
	}
	}
#else
	while (1)
	{
#if 1
		for (point = 0; point < 1; point++)
		{
			p0 = shape[point];
			p1 = shape[point+1];
			p0.x += pos; p0.y += pos;
			p1.x += pos; p1.y += pos;
			draw_line(&p0, &p1);
		}

		pos += delta * 0.001;
		if (pos > 0.3)
			{ pos = 0.3; delta = -1; }
		else if (pos < -0.3)
			{ pos = -0.3; delta = 1; }
#else
#endif
	}
#endif

	write(fd, reset_vals, 3 * sizeof(unsigned short));
	close(fd);

	return 0;
}

